from ..service import ReceivingService, ThreadService
from ..wire import RawWire

class SaveRaw(ThreadService, ReceivingService):
    input_wire_class=RawWire
    def __init__(self, filename=None):
        super(SaveRaw, self).__init__()
        
        if filename is None:
            filename="data.raw"
        
        self.outputfile=filename
        
        self.output=open(self.outputfile, 'wb')

    def process(self, data):
        self.output.write(data)

    def cleanup(self):
        self.output.close()
        super(SaveRaw, self).cleanup()



import sys
import pickle

from ..service import ReceivingService, ThreadService
from ..wire import PulseWire

class Count(ThreadService, ReceivingService):
    input_wire_class=PulseWire
    def __init__(self, outfile="count"):
        super(Count, self).__init__()
        self.all_pulses=[]
        self.outfile=outfile
        
    def process(self, data):
        if len(data)>0:
            self.all_pulses.extend(data)
        total=len(self.all_pulses)
        cps=0
        cps10=0
        dt=0
        if total>0:
            dt=self.all_pulses[-1][0]-self.all_pulses[0][0]
            if dt>0:
                cps=total/dt
        if total>101:
            dt10=self.all_pulses[-1][0]-self.all_pulses[-100][0]
            if dt10>0:
                cps10=100./dt10
        else:
            cps10=cps
                
        message="time: {0:6.3g} | counts: {1:5.3e} | average cps: {2:5.2f} | current cps: {3:5.2f}".format(dt, total, cps, cps10)
        self.print_message(message)
        
    def cleanup(self):
        if self.outfile is not None:
            f=open(self.outfile+".pkl","wb")
            pickle.dump(self.all_pulses,f)
            f.close()
        super(Count, self).cleanup()



import numpy
import time
from ..service import SourceService, ThreadService
from ..wire import RawWire

class Noise(ThreadService, SourceService):
    output_wire_class=RawWire
    def __init__(self, frames_per_buffer=2048):
        super(Noise, self).__init__()
        self.CHANNELS=1
        self.RATE=48000
        self.FORMAT="float32"
        self.t0=time.time()
        self.frames_per_buffer=frames_per_buffer

    def output_protocol(self, wire):
        super(Noise, self).output_protocol(wire)
        wire.CHANNELS=self.CHANNELS
        wire.RATE=self.RATE
        wire.FORMAT=self.FORMAT
    
    def process(self, data=None):
        time.sleep(self.frames_per_buffer/(1.*self.RATE))
        t=time.time()
        dt=t-self.t0
        self.t0=t
        
        n=int(numpy.floor(dt*self.RATE))
        data=numpy.random.random(n)
        return data.astype("float32").tobytes()




from ..service import ReceivingService, SourceService, ThreadService
from ..wire import RawWire, FloatWire

import numpy

class Identity(ThreadService, SourceService, ReceivingService):
    input_wire_class=RawWire
    output_wire_class=RawWire
    
    def output_protocol(self, wire):
        super(Identity, self).output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT=self.input_wire.FORMAT

    def process(self, data):
        return data

class Raw2Numpy(ThreadService, SourceService, ReceivingService):
    input_wire_class=RawWire
    output_wire_class=FloatWire

    def output_protocol(self, wire):
        super(Raw2Numpy, self).output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT=self.input_wire.FORMAT

    def process(self, data):
        return numpy.frombuffer(data,dtype=self.input_wire.FORMAT)

class Raw2Float(ThreadService, SourceService, ReceivingService):
    input_wire_class=RawWire
    output_wire_class=FloatWire

    def output_protocol(self, wire):
        super(Raw2Float, self).output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.RATE=self.input_wire.RATE
        wire.FORMAT="float32"

    def process(self, data):
        if self.input_wire.FORMAT=="float32":
          return numpy.frombuffer(data,dtype=self.input_wire.FORMAT)
        elif self.input_wire.FORMAT=="int16":
          data=numpy.frombuffer(data,dtype=self.input_wire.FORMAT)
          return data.astype("float32")/32768
        else:
          self.print_message("unknown data format in wire")
          return None

class DownSampleMaxed(ThreadService, SourceService, ReceivingService):
    input_wire_class=FloatWire
    output_wire_class=FloatWire

    def __init__(self, factor=8):
        super(DownSampleMaxed, self).__init__()
        self.factor=factor

    def output_protocol(self, wire):
        super(DownSampleMaxed, self).output_protocol(wire)
        wire.CHANNELS=self.input_wire.CHANNELS
        wire.FORMAT=self.input_wire.FORMAT
        wire.RATE=self.input_wire.RATE/self.factor

    def process(self, data):
        return numpy.max(data.reshape(-1, self.factor),axis=1)



import numpy
from ..service import ReceivingService, ThreadService
from ..wire import RawWire

class SoundCardPlay(ThreadService, ReceivingService):

    @staticmethod
    def devices():
        """ return dict with names and ids of sound devices for playback (aka speakers) """
        result=dict()
        try:
            import soundcard
        except Exception as ex:
            raise Exception( "import error: {0}".format(str(ex)))

        for m in soundcard.all_speakers():
          result[m.name]=m.id
        return result

    input_wire_class=RawWire

    def __init__(self, frames_per_buffer=2048, output_device_index=None, output_device_name=""):
        self.frames_per_buffer=frames_per_buffer
        self.output_device_index=output_device_index
        self.output_device_name=output_device_name
        super(SoundCardPlay, self).__init__()

    def connect_input(self,service):
        super(SoundCardPlay, self).connect_input(service)
        if self.input_wire.FORMAT != "float32":
            raise Exception("SoundCard playback only supports float32 format")

    def _process(self):
        global soundcard
        try:
            import soundcard
        except Exception as ex:
            self.print_message( "import error: {0}".format(str(ex)))
            
        speaker=soundcard.get_speaker(self.output_device_index or self.output_device_name)
        self.print_message( f"opening {str(speaker)} for audio output")

        with speaker.player(samplerate=self.input_wire.RATE, blocksize=self.frames_per_buffer) as player:
            while not self.stopped:
                data=self.receive_input()
                if data is None:
                    self.stopped=True
                else:
                    data=numpy.frombuffer(data,dtype=self.input_wire.FORMAT)

                if not self.stopped:
                    try:
                        player.play(data)
                    except Exception as ex:
                        self.print_message( "error: {0}".format(str(ex)))
                        self.stopped=True
        self.stopped=True




from ..service import ThreadService, ReceivingService
from ..wire import FloatWire, HistogramWire

import numpy
import time

class Monitor(ThreadService, ReceivingService):
    input_wire_class=FloatWire

    def __init__(self, window=5, vmin=-0.01, vmax=1.1, outfile=None):                
        self.window=window
        self.vmin=vmin
        self.vmax=vmax
        self.outfile=outfile
        self.do_update=True

        super(Monitor, self).__init__()
        self.thread.daemon=True

    def update_plot(self,nframe):

        if self.do_update==False:
            return self.plot

        data=[]
        while True:
            _data=self.receive_input(False)
            if _data is None:
                break
            data.append(_data)

        if data:
            data=numpy.concatenate(data)
        else:
            return self.plot

        if self.stopped:
            self.do_update=False
            if self.outfile is not None:
              self.fig.savefig(self.outfile+'.png')
            self.done=True
            return self.plot
        
        while len(data)>0:
          
            n=min(len(data),len(self.plotdata)-self.nplot)
            
            self.plotdata[self.nplot:self.nplot+n]=data[:n]
            data=data[n:]
            self.nplot=(self.nplot+n) % len(self.plotdata)
                            
        self.plot[0].set_ydata(self.plotdata)
        return self.plot
  
    def start_process(self):
        global FuncAnimation, pyplot
        try:
            from matplotlib.animation import FuncAnimation
            from matplotlib import pyplot    
        except Exception as ex:
            self.print_message("import error: {0}".format(str(ex)))

        self.plotdata=numpy.zeros(int(self.input_wire.RATE*self.window), dtype=self.input_wire.FORMAT)
        self.plotx=numpy.arange(int(self.input_wire.RATE*self.window))/float(self.input_wire.RATE)

        self.fig, self.ax = pyplot.subplots()
        self.fig.canvas.manager.set_window_title("GammaLab Monitor")

        self.ax.cla()
        self.plot=self.ax.plot(self.plotx,self.plotdata)
        self.ax.set_ylim(self.vmin,self.vmax)
        self.ax.set_xlabel("time (s)")
        self.ax.set_ylabel("level")
      
        self.nplot=0

        ani = FuncAnimation(self.fig, self.update_plot, interval=250, repeat=False, blit=True)

        pyplot.show(block=True)

        if self.outfile is not None:
            self.fig.savefig(self.outfile+'.png')

        self.cleanup()
        
    def close(self):
        # deamon process
        self.stop()

        while not self.done:
            time.sleep(0.1)
      

class PlotHistogram(ThreadService, ReceivingService):
    input_wire_class=HistogramWire
  
    def __init__(self, xmin=0, xmax=1., log=True, 
                    error_bars=True, outfile="histogram"):        
        self.xmin=xmin
        self.xmax=xmax
        self.log=log
        self.error_bars=error_bars
        self.outfile=outfile
        self.ymax=1.
        self.do_update=True

        super(PlotHistogram, self).__init__()
        self.thread.daemon=True

    def update_plot(self,nframe):
        
        data=None
        while True:
            _data=self.receive_input(False)
            if _data is None:
                break
            data=_data

        if self.stopped or not self.do_update or data is None:
            if self.stopped:
                self.do_update=False
                if self.outfile is not None:
                  self.fig.savefig(self.outfile+'.png')
                self.done=True
          
            if self.error_bars:
                return self._line,self._top,self._bot
            else:
                return self._line,

        hist=data["hist"]
        bins=data["bins"]

        return self.update_histogram_plot(hist,bins)

    def update_histogram_plot(self, hist, bins):
        y=numpy.zeros(2*len(hist))
        y[::2]=hist
        y[1::2]=hist
        yerr=y**0.5
        self._line.set_ydata(y)
        self._update_ylim(y.max())
        if self.error_bars:
            self._top.set_ydata(numpy.maximum(y+yerr,1))
            self._bot.set_ydata(y-yerr)
                
        if self.error_bars:
            return self._line,self._top,self._bot
        else:
            return self._line,

    def _update_ylim(self,ymax):
        if self.log:
            _ymax=ymax
            if _ymax>self.ymax:
                self.ymax=10*_ymax
                self.ax.set_ylim(0.5,2*self.ymax)
        else:
            _ymax=ymax
            if _ymax>self.ymax:
                self.ymax=1.5*ymax
                self.ax.set_ylim(0,1.3*self.ymax)      

    def _histogram_plot(self):
        self.ax.cla()
        nchannels=self.input_wire.nchannels
        vmin=self.input_wire.vmin
        vmax=self.input_wire.vmax
        hist,bins=numpy.histogram([], bins=nchannels, range=(vmin,vmax))
        x=numpy.zeros(2*nchannels)
        y=numpy.zeros(2*nchannels)
        x[::2]=bins[:-1]
        x[1::2]=bins[1:]
        y[::2]=hist
        y[1::2]=hist
        if self.error_bars:
            yerr=y**0.5
            self._top,=self.ax.plot(x,numpy.maximum(y+yerr,1),":",c="tab:orange", zorder=0)
            self._bot,=self.ax.plot(x,y-yerr,":",c="tab:orange",zorder=5)
        if self.log:
            self._line,=self.ax.semilogy(x,y+0.1, lw=2, zorder=10)
        else:
            self._line,=self.ax.plot(x,y,lw=2, zorder=10)

        self.ax.set_ylabel("counts")
        self.ax.set_xlabel(f"energy ({self.input_wire.unit})") # take label from wire?
        self.ax.set_xlim(self.xmin,self.xmax)
        self._update_ylim(max(y.max(),50))
      
    def start_process(self):
        global FuncAnimation, pyplot
        try:
            from matplotlib.animation import FuncAnimation
            from matplotlib import pyplot    
        except Exception as ex:
            self.print_message("import error: {0}".format(str(ex)))

        self.fig, self.ax = pyplot.subplots()
        self.fig.canvas.manager.set_window_title("GammaLab Histogram")
      
        self._histogram_plot()

        ani = FuncAnimation(self.fig, self.update_plot, interval=250, repeat=False, blit=True)

        pyplot.show(block=True)
                
        if self.outfile is not None:
            self.fig.savefig(self.outfile+'.png')

        self.cleanup()

    def close(self):
        # deamon process
        self.stop()

        while not self.done:
            time.sleep(0.1)




#!/usr/bin/env python
"""
gammalab-monitor plots raw detector (soundcard) values. 
"""
import argparse
from datetime import datetime

try:
  import matplotlib
  matplotlib.use('TkAgg')
except:
  pass

from gammalab import main
from gammalab.acquisition import SoundCard, FileReplay
from gammalab.backend import Monitor
from gammalab.backend import SaveRaw
from gammalab.backend import SoundCardPlay
from gammalab.transform import Raw2Float, DownSampleMaxed

def run(input_device_name="", runtime=None, list_input_devices=False, raw_output_file="",
        list_output_devices=False, output_device_name="", inputfile=None):
    if list_input_devices:
      for name, id_ in SoundCard.devices().items():
        print("{0}: {1}".format(name, id_))
      exit(0)
    if list_output_devices:
      for name, id_ in SoundCardPlay.devices().items():
        print("{0}: {1}".format(name, id_))
      exit(0)

    if inputfile is not None:
        source=FileReplay(inputfile)
    else:
        source=SoundCard(input_device_name=input_device_name)

    monitor=Monitor()
    convert=Raw2Float()
    downsample=DownSampleMaxed(factor=8)
        
    source.plugs_into(convert)
    convert.plugs_into(downsample)
    downsample.plugs_into(monitor)

    if raw_output_file!="":
        output_file=raw_output_file+"."+datetime.utcnow().strftime("%Y%m%d%H%M%S")
        save=SaveRaw(output_file)
        source.plugs_into(save)

    if output_device_name!="":
        playback=SoundCardPlay(output_device_name=output_device_name)
        source.plugs_into(playback)


    main(runtime)

def new_argument_parser():
    "Parse command line arguments"
    parser = argparse.ArgumentParser( formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                      description=__doc__)
    parser.add_argument(
        '--input_device_name',
        dest='input_device_name',
        default="",
        type=str,
        help='select input device by (fuzzy matched) name',
    )
    parser.add_argument(
        '--runtime',
        dest='runtime',
        default=None,
        type=float,
        help='runtime in seconds',
    )
    parser.add_argument(
        '--list_input_devices',
        dest='list_input_devices',
        action="store_true",
        default=False,
        help='list all input devices',
    )
    parser.add_argument(
        '--list_output_devices',
        dest='list_output_devices',
        action="store_true",
        default=False,
        help='list output soundcard devices',
    )
    parser.add_argument(
        '--raw_ouput_file',
        dest='raw_output_file',
        default="",
        type=str,
        help='optionally record raw data stream to provided filename (for later analysis) ',
    )
    parser.add_argument(
        '--output_device_name',
        dest='output_device_name',
        default="",
        type=str,
        help='optional output device for playback (fuzzy matched by name)',
    )
    parser.add_argument(
        '--infile',
        dest='inputfile',
        default=None,
        help='Optionally playback from input file',
    ) 
    return parser.parse_args()

if __name__=="__main__":
    args = new_argument_parser()
    run(**vars(args))



from ..service import SourceService, ThreadService
from ..wire import RawWire

class SoundCard(ThreadService, SourceService):
    output_wire_class=RawWire

    @staticmethod
    def devices():
        """ return dict with names and ids of sound devices for acquisition (aka microphones) """
        try:
            import soundcard
        except Exception as ex:
            raise Exception( "import error: {0}".format(str(ex)))
        result=dict()
        for m in soundcard.all_microphones():
          result[m.name]=m.id
        return result
  
    def __init__(self, frames_per_buffer=2048, input_device_index=None, input_device_name="",
                 sample_rate=48000, sample_format="float32"):
        self.CHANNELS=1
        self.RATE=sample_rate
        if sample_format != "float32":
          raise Exception("SoundCard only supports float32 sample_format")
        self.FORMAT=sample_format
        self.frames_per_buffer=frames_per_buffer
        self.input_device_index=input_device_index
        self.input_device_name=input_device_name
        super(SoundCard, self).__init__()

    def _process(self):
        global soundcard
        try:
            import soundcard
        except Exception as ex:
            self.print_message( "import error: {0}".format(str(ex)))

        mic=soundcard.get_microphone(self.input_device_index or self.input_device_name)
        with mic.recorder(samplerate=self.RATE, blocksize=self.frames_per_buffer, channels=self.CHANNELS) as recorder:
            while not self.stopped:    
                try:
                    data=recorder.record(self.frames_per_buffer)
                except Exception as ex:
                    self.print_message( "error: {0}".format(str(ex)))
                    self.stopped=True
                if len(data)==0:
                    self.print_message("no data")
                    self.stopped=True
                else:
                    data=data.astype("float32").tobytes()
    
                if (not self.stopped and
                    data is not None):
                    self.send_output(data)
                
            self.send_output(None)
        self.stopped=True

    def output_protocol(self, wire):
        super(SoundCard, self).output_protocol(wire)
        wire.CHANNELS=self.CHANNELS
        wire.RATE=self.RATE
        wire.FORMAT=self.FORMAT



from ..service import SourceService
from ..wire import RawWire

try:
    import pyaudio
    HAS_PYAUDIO=True

    pyaudio_format=dict(int16=pyaudio.paInt16, float32=pyaudio.paFloat32)
except ImportError:
    HAS_PYAUDIO=False


class PyAudio(SourceService):
    output_wire_class=RawWire
    def __init__(self, frames_per_buffer=2048, input_device_index=None, 
                 sample_rate=48000, sample_format="float32"):
        if not HAS_PYAUDIO:
          raise Exception("pyaudio module not or not correctly installed")
        self.CHANNELS=1
        self.RATE=sample_rate
        self.FORMAT=sample_format
        self.pyaudio=pyaudio.PyAudio()
        self.recorder=None
        self.frames_per_buffer=frames_per_buffer
        self.input_device_index=input_device_index
        super(PyAudio, self).__init__()
        
    def output_protocol(self, wire):
        super(PyAudio, self).output_protocol(wire)      
        wire.CHANNELS=self.CHANNELS
        wire.RATE=self.RATE
        wire.FORMAT=self.FORMAT
        
    def _callback(self, in_data, frame_count, time_info, status):
        self.send_output(in_data)
        if status!=0:
            self.print_message("status = {0}".format(status))
        return (None, pyaudio.paContinue)

    def start(self):
        self.recorder = self.pyaudio.open(
                format=pyaudio_format[self.FORMAT],
                channels=self.CHANNELS,
                rate=self.RATE,
                frames_per_buffer=self.frames_per_buffer,
                input=True,
                stream_callback=self._callback,
                input_device_index=self.input_device_index
                )
                
    def stop(self):
        if self.recorder:
            self.recorder.stop_stream()

    def close(self):
        if self.recorder:
            self.recorder.stop_stream()
            self.recorder.close()

        self.pyaudio.terminate()



#!/usr/bin/env python
"""
gammalab-calibration-cal determines calibration coeff for a polynomial calibration curve. 
"""


import argparse
import numpy

def calibration_pair(s):
    return list(map(float, s.split(",")))

def new_argument_parser():

    parser = argparse.ArgumentParser(description='Determine scale and drift parameters from calibration data')
    parser.add_argument(
        dest='data', 
        default=None, 
        nargs="*", 
        type=calibration_pair,
        help='calibration data: a list of space sperated <raw value>,<energy> pairs, e.g.: 0.1,100 0.5,500.'
    )

    parser.add_argument(
        '--order',
        dest='order',
        default=2,
        type=int,
        help='order of calibration polynomial fit',
    )
  
    return parser.parse_args()

def run(data=None, order=2):
    if data is None:
        raise Exception("no data, provide --data ..")
    
    if order>len(data):
        order=len(data)
    
    data=numpy.array(data)
    x=data[:,0]
    y=data[:,1]
    
    a=numpy.zeros((len(x), order))
    
    for i in range(order):
        a[:,i]=x**(i+1)
    
    p=numpy.linalg.pinv(a)
    
    coeff=p.dot(y)
    
    print(f"raw coefficents: {coeff}")
    print(f"the scale is: {coeff[0]}")
    if order>=2:
        print(f"the drift is: {coeff[1]/coeff[0]}")
    
    

if __name__=="__main__":
    args=new_argument_parser()
    run(**vars(args))
